/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2014, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPattern.cpp
    \brief    neighbouring pixel access classes
*/

#include "TComPic.h"
#include "TComPattern.h"
#include "TComDataCU.h"
#include "TComTU.h"
#include "Debug.h"
#include "TComPrediction.h"

//! \ingroup TLibCommon
//! \{

// Forward declarations

/// padding of unavailable reference samples for intra prediction
#if RExt__O0043_BEST_EFFORT_DECODING
Void fillReferenceSamples( const Int bitDepth, const Int bitDepthDelta, TComDataCU* pcCU, const Pel* piRoiOrigin, Pel* piAdiTemp, const Bool* bNeighborFlags,
#else
Void fillReferenceSamples( const Int bitDepth, TComDataCU* pcCU, const Pel* piRoiOrigin, Pel* piAdiTemp, const Bool* bNeighborFlags,
#endif
                           const Int iNumIntraNeighbor, const Int unitWidth, const Int unitHeight, const Int iAboveUnits, const Int iLeftUnits,
                           const UInt uiCuWidth, const UInt uiCuHeight, const UInt uiWidth, const UInt uiHeight, const Int iPicStride,
                           const ChannelType chType, const ChromaFormat chFmt );

/// constrained intra prediction
Bool  isAboveLeftAvailable  ( TComDataCU* pcCU, UInt uiPartIdxLT );
Int   isAboveAvailable      ( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxRT, Bool* bValidFlags );
Int   isLeftAvailable       ( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxLB, Bool* bValidFlags );
Int   isAboveRightAvailable ( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxRT, Bool* bValidFlags );
Int   isBelowLeftAvailable  ( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxLB, Bool* bValidFlags );


// ====================================================================================================================
// Public member functions (TComPatternParam)
// ====================================================================================================================

/** \param  piTexture     pixel data
 \param  iRoiWidth     pattern width
 \param  iRoiHeight    pattern height
 \param  iStride       buffer stride
 \param  iOffsetLeft   neighbour offset (left)
 \param  iOffsetRight  neighbour offset (right)
 \param  iOffsetAbove  neighbour offset (above)
 \param  iOffsetBottom neighbour offset (bottom)
 */
Void TComPatternParam::setPatternParamPel ( Pel* piTexture,
                                           Int iRoiWidth,
                                           Int iRoiHeight,
                                           Int iStride
                                           )
{
  m_piROIOrigin    = piTexture;
  m_iROIWidth       = iRoiWidth;
  m_iROIHeight      = iRoiHeight;
  m_iPatternStride  = iStride;
}

// ====================================================================================================================
// Public member functions (TComPattern)
// ====================================================================================================================

Void TComPattern::initPattern (Pel* piY,
                               Int iRoiWidth,
                               Int iRoiHeight,
                               Int iStride)
{
  m_cPatternY. setPatternParamPel( piY,  iRoiWidth, iRoiHeight, iStride);
}


// NOTE: RExt - this has been kept in this C++ file to allow easier tracking/comparison against HM.
Void TComPrediction::initAdiPatternChType( TComTU &rTu, Bool& bAbove, Bool& bLeft, const ComponentID compID, const Bool bFilterRefSamples DEBUG_STRING_FN_DECLARE(sDebug))
{
  // NOTE: RExt - This function has been modified for increased flexibility as part of the Square TU 4:2:2 implementation
  const ChannelType chType    = toChannelType(compID);

  TComDataCU *pcCU=rTu.getCU();
  const UInt uiZorderIdxInPart=rTu.GetAbsPartIdxTU();
  const UInt uiTuWidth        = rTu.getRect(compID).width;
  const UInt uiTuHeight       = rTu.getRect(compID).height;
  const UInt uiTuWidth2       = uiTuWidth  << 1;
  const UInt uiTuHeight2      = uiTuHeight << 1;

  const Int  iBaseUnitSize    = g_uiMaxCUWidth >> g_uiMaxCUDepth;
  const Int  iUnitWidth       = iBaseUnitSize  >> pcCU->getPic()->getPicYuvRec()->getComponentScaleX(compID);
  const Int  iUnitHeight      = iBaseUnitSize  >> pcCU->getPic()->getPicYuvRec()->getComponentScaleY(compID);
  const Int  iTUWidthInUnits  = uiTuWidth  / iUnitWidth;
  const Int  iTUHeightInUnits = uiTuHeight / iUnitHeight;
  const Int  iAboveUnits      = iTUWidthInUnits  << 1;
  const Int  iLeftUnits       = iTUHeightInUnits << 1;

  assert(iTUHeightInUnits > 0 && iTUWidthInUnits > 0);

  const Int  iPartIdxStride   = pcCU->getPic()->getNumPartInWidth(); //NOTE: RExt - despite the name "TComPic::getNumPartInWidth", this gets the number of parts in an **LCU** width, not the picture width
  const UInt uiPartIdxLT      = pcCU->getZorderIdxInCU() + uiZorderIdxInPart;
  const UInt uiPartIdxRT      = g_auiRasterToZscan[ g_auiZscanToRaster[ uiPartIdxLT ] +   iTUWidthInUnits  - 1                   ];
  const UInt uiPartIdxLB      = g_auiRasterToZscan[ g_auiZscanToRaster[ uiPartIdxLT ] + ((iTUHeightInUnits - 1) * iPartIdxStride)];

  Int   iPicStride = pcCU->getPic()->getStride(compID);
  Bool  bNeighborFlags[4 * MAX_NUM_SPU_W + 1];
  Int   iNumIntraNeighbor = 0;

  bNeighborFlags[iLeftUnits] = isAboveLeftAvailable( pcCU, uiPartIdxLT );
  iNumIntraNeighbor += bNeighborFlags[iLeftUnits] ? 1 : 0;
  iNumIntraNeighbor  += isAboveAvailable     ( pcCU, uiPartIdxLT, uiPartIdxRT, (bNeighborFlags + iLeftUnits + 1)                    );
  iNumIntraNeighbor  += isAboveRightAvailable( pcCU, uiPartIdxLT, uiPartIdxRT, (bNeighborFlags + iLeftUnits + 1 + iTUWidthInUnits ) );
  iNumIntraNeighbor  += isLeftAvailable      ( pcCU, uiPartIdxLT, uiPartIdxLB, (bNeighborFlags + iLeftUnits - 1)                    );
  iNumIntraNeighbor  += isBelowLeftAvailable ( pcCU, uiPartIdxLT, uiPartIdxLB, (bNeighborFlags + iLeftUnits - 1 - iTUHeightInUnits) );

  bAbove = true;
  bLeft  = true;

  const ChromaFormat chFmt       = rTu.GetChromaFormat();
  const UInt         uiROIWidth  = uiTuWidth2+1;
  const UInt         uiROIHeight = uiTuHeight2+1;

  assert(uiROIWidth*uiROIHeight <= m_iYuvExtSize);

#ifdef DEBUG_STRING
  std::stringstream ss(stringstream::out);
#endif

  {
    Pel *piAdiTemp   = m_piYuvExt[compID][PRED_BUF_UNFILTERED];
    Pel *piRoiOrigin = pcCU->getPic()->getPicYuvRec()->getAddr(compID, pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiZorderIdxInPart);
#if RExt__O0043_BEST_EFFORT_DECODING
    fillReferenceSamples (g_bitDepthInStream[chType], g_bitDepthInStream[chType] - g_bitDepth[chType], pcCU, piRoiOrigin, piAdiTemp, bNeighborFlags, iNumIntraNeighbor,  iUnitWidth, iUnitHeight, iAboveUnits, iLeftUnits,
#else
    fillReferenceSamples (g_bitDepth[chType], pcCU, piRoiOrigin, piAdiTemp, bNeighborFlags, iNumIntraNeighbor,  iUnitWidth, iUnitHeight, iAboveUnits, iLeftUnits,
#endif
                          uiTuWidth, uiTuHeight, uiROIWidth, uiROIHeight, iPicStride, toChannelType(compID), chFmt);


#ifdef DEBUG_STRING
    if (DebugOptionList::DebugString_Pred.getInt()&DebugStringGetPredModeMask(MODE_INTRA))
    {
      ss << "###: generating Ref Samples for channel " << compID << " and " << rTu.getRect(compID).width << " x " << rTu.getRect(compID).height << "\n";
      for (UInt y=0; y<uiROIHeight; y++)
      {
        ss << "###: - ";
        for (UInt x=0; x<uiROIWidth; x++)
        {
          if (x==0 || y==0)
            ss << piAdiTemp[y*uiROIWidth + x] << ", ";
//          if (x%16==15) ss << "\nPart size: ~ ";
        }
        ss << "\n";
      }
    }
#endif

    if (bFilterRefSamples)
    {
      // generate filtered intra prediction samples

            Int          stride    = uiROIWidth;
      const Pel         *piSrcPtr  = piAdiTemp                             + (stride * uiTuHeight2); // bottom left
            Pel         *piDestPtr = m_piYuvExt[compID][PRED_BUF_FILTERED] + (stride * uiTuHeight2); // bottom left

      //------------------------------------------------

      Bool useStrongIntraSmoothing = isLuma(chType) && pcCU->getSlice()->getSPS()->getUseStrongIntraSmoothing();

      const Pel bottomLeft = piAdiTemp[stride * uiTuHeight2];
      const Pel topLeft    = piAdiTemp[0];
      const Pel topRight   = piAdiTemp[uiTuWidth2];

      if (useStrongIntraSmoothing)
      {
#if RExt__O0043_BEST_EFFORT_DECODING
        const Int  threshold     = 1 << (g_bitDepthInStream[chType] - 5);
#else
        const Int  threshold     = 1 << (g_bitDepth[chType] - 5);
#endif
        const Bool bilinearLeft  = abs((bottomLeft + topLeft ) - (2 * piAdiTemp[stride * uiTuHeight])) < threshold; //difference between the
        const Bool bilinearAbove = abs((topLeft    + topRight) - (2 * piAdiTemp[         uiTuWidth ])) < threshold; //ends and the middle
        if ((uiTuWidth < 32) || (!bilinearLeft) || (!bilinearAbove))
          useStrongIntraSmoothing = false;
      }

      *piDestPtr = *piSrcPtr; // bottom left is not filtered
      piDestPtr -= stride;
      piSrcPtr  -= stride;

      //------------------------------------------------

      //left column (bottom to top)

      if (useStrongIntraSmoothing)
      {
        const Int shift = g_aucConvertToBit[uiTuHeight] + 3; //log2(uiTuHeight2)

        for(UInt i=1; i<uiTuHeight2; i++, piDestPtr-=stride)
        {
          *piDestPtr = (((uiTuHeight2 - i) * bottomLeft) + (i * topLeft) + uiTuHeight) >> shift;
        }

        piSrcPtr -= stride * (uiTuHeight2 - 1);
      }
      else
      {
        for(UInt i=1; i<uiTuHeight2; i++, piDestPtr-=stride, piSrcPtr-=stride)
        {
          *piDestPtr = ( piSrcPtr[stride] + 2*piSrcPtr[0] + piSrcPtr[-stride] + 2 ) >> 2;
        }
      }

      //------------------------------------------------

      //top-left

      if (useStrongIntraSmoothing)
      {
        *piDestPtr = piSrcPtr[0];
      }
      else
      {
        *piDestPtr = ( piSrcPtr[stride] + 2*piSrcPtr[0] + piSrcPtr[1] + 2 ) >> 2;
      }
      piDestPtr += 1;
      piSrcPtr  += 1;

      //------------------------------------------------

      //top row (left-to-right)

      if (useStrongIntraSmoothing)
      {
        const Int shift = g_aucConvertToBit[uiTuWidth] + 3; //log2(uiTuWidth2)

        for(UInt i=1; i<uiTuWidth2; i++, piDestPtr++)
        {
          *piDestPtr = (((uiTuWidth2 - i) * topLeft) + (i * topRight) + uiTuWidth) >> shift;
        }

        piSrcPtr += uiTuWidth2 - 1;
      }
      else
      {
        for(UInt i=1; i<uiTuWidth2; i++, piDestPtr++, piSrcPtr++)
        {
          *piDestPtr = ( piSrcPtr[1] + 2*piSrcPtr[0] + piSrcPtr[-1] + 2 ) >> 2;
        }
      }

      //------------------------------------------------

      *piDestPtr=*piSrcPtr; // far right is not filtered

#ifdef DEBUG_STRING
    if (DebugOptionList::DebugString_Pred.getInt()&DebugStringGetPredModeMask(MODE_INTRA))
    {
      ss << "###: filtered result for channel " << compID <<"\n";
      for (UInt y=0; y<uiROIHeight; y++)
      {
        ss << "###: - ";
        for (UInt x=0; x<uiROIWidth; x++)
        {
          if (x==0 || y==0)
            ss << m_piYuvExt[compID][PRED_BUF_FILTERED][y*uiROIWidth + x] << ", ";
//          if (x%16==15) ss << "\nPart size: ~ ";
        }
        ss << "\n";
      }
    }
#endif


    }
  }
  DEBUG_STRING_APPEND(sDebug, ss.str())
}

#if RExt__O0043_BEST_EFFORT_DECODING
Void fillReferenceSamples( const Int bitDepth, const Int bitDepthDelta, TComDataCU* pcCU, const Pel* piRoiOrigin, Pel* piAdiTemp, const Bool* bNeighborFlags,
#else
Void fillReferenceSamples( const Int bitDepth, TComDataCU* pcCU, const Pel* piRoiOrigin, Pel* piAdiTemp, const Bool* bNeighborFlags,
#endif
                           const Int iNumIntraNeighbor, const Int unitWidth, const Int unitHeight, const Int iAboveUnits, const Int iLeftUnits,
                           const UInt uiCuWidth, const UInt uiCuHeight, const UInt uiWidth, const UInt uiHeight, const Int iPicStride,
                           const ChannelType chType, const ChromaFormat chFmt )
{
  // NOTE: RExt - This function has been modified for increased flexibility as part of the Square TU 4:2:2 implementation
  const Pel* piRoiTemp;
  Int  i, j;
  Int  iDCValue = 1 << (bitDepth - 1);
  const Int iTotalUnits = iAboveUnits + iLeftUnits + 1; //+1 for top-left

  if (iNumIntraNeighbor == 0)
  {
    // Fill border with DC value
    for (i=0; i<uiWidth; i++)
    {
      piAdiTemp[i] = iDCValue;
    }
    for (i=1; i<uiHeight; i++)
    {
      piAdiTemp[i*uiWidth] = iDCValue;
    }
  }
  else if (iNumIntraNeighbor == iTotalUnits)
  {
    // Fill top-left border and top and top right with rec. samples
    piRoiTemp = piRoiOrigin - iPicStride - 1;

    for (i=0; i<uiWidth; i++)
    {
#if RExt__O0043_BEST_EFFORT_DECODING
      piAdiTemp[i] = piRoiTemp[i] << bitDepthDelta;
#else
      piAdiTemp[i] = piRoiTemp[i];
#endif
    }

    // Fill left and below left border with rec. samples
    piRoiTemp = piRoiOrigin - 1;

    for (i=1; i<uiHeight; i++)
    {
#if RExt__O0043_BEST_EFFORT_DECODING
      piAdiTemp[i*uiWidth] = (*(piRoiTemp)) << bitDepthDelta;
#else
      piAdiTemp[i*uiWidth] = *(piRoiTemp);
#endif
      piRoiTemp += iPicStride;
    }
  }
  else // reference samples are partially available
  {
    // all above units have "unitWidth" samples each, all left/below-left units have "unitHeight" samples each
    const Int  iTotalSamples = (iLeftUnits * unitHeight) + ((iAboveUnits + 1) * unitWidth);
    Pel  piAdiLine[5 * MAX_CU_SIZE];
    Pel  *piAdiLineTemp;
    const Bool *pbNeighborFlags;


    // Initialize
    for (i=0; i<iTotalSamples; i++)
    {
      piAdiLine[i] = iDCValue;
    }

    // Fill top-left sample
    piRoiTemp = piRoiOrigin - iPicStride - 1;
    piAdiLineTemp = piAdiLine + (iLeftUnits * unitHeight);
    pbNeighborFlags = bNeighborFlags + iLeftUnits;
    if (*pbNeighborFlags)
    {
#if RExt__O0043_BEST_EFFORT_DECODING
      Pel topLeftVal=piRoiTemp[0] << bitDepthDelta;
#else
      Pel topLeftVal=piRoiTemp[0];
#endif
      for (i=0; i<unitWidth; i++)
      {
        piAdiLineTemp[i] = topLeftVal;
      }
    }

    // Fill left & below-left samples (downwards)
    piRoiTemp += iPicStride;
    piAdiLineTemp--;
    pbNeighborFlags--;

    for (j=0; j<iLeftUnits; j++)
    {
      if (*pbNeighborFlags)
      {
        for (i=0; i<unitHeight; i++)
        {
#if RExt__O0043_BEST_EFFORT_DECODING
          piAdiLineTemp[-i] = piRoiTemp[i*iPicStride] << bitDepthDelta;
#else
          piAdiLineTemp[-i] = piRoiTemp[i*iPicStride];
#endif
        }
      }
      piRoiTemp += unitHeight*iPicStride;
      piAdiLineTemp -= unitHeight;
      pbNeighborFlags--;
    }

    // Fill above & above-right samples (left-to-right) (each unit has "unitWidth" samples)
    piRoiTemp = piRoiOrigin - iPicStride;
    // offset line buffer by iNumUints2*unitHeight (for left/below-left) + unitWidth (for above-left)
    piAdiLineTemp = piAdiLine + (iLeftUnits * unitHeight) + unitWidth;
    pbNeighborFlags = bNeighborFlags + iLeftUnits + 1;
    for (j=0; j<iAboveUnits; j++)
    {
      if (*pbNeighborFlags)
      {
        for (i=0; i<unitWidth; i++)
        {
#if RExt__O0043_BEST_EFFORT_DECODING
          piAdiLineTemp[i] = piRoiTemp[i] << bitDepthDelta;
#else
          piAdiLineTemp[i] = piRoiTemp[i];
#endif
        }
      }
      piRoiTemp += unitWidth;
      piAdiLineTemp += unitWidth;
      pbNeighborFlags++;
    }

    // Pad reference samples when necessary
    Int iCurrJnit = 0;
    Pel  *piAdiLineCur   = piAdiLine;
    const UInt piAdiLineTopRowOffset = iLeftUnits * (unitHeight - unitWidth);

    if (!bNeighborFlags[0])
    {
      // very bottom unit of bottom-left; at least one unit will be valid.
      {
        Int   iNext = 1;
        while (iNext < iTotalUnits && !bNeighborFlags[iNext])
        {
          iNext++;
        }
        Pel *piAdiLineNext = piAdiLine + ((iNext < iLeftUnits) ? (iNext * unitHeight) : (piAdiLineTopRowOffset + (iNext * unitWidth)));
        const Pel refSample = *piAdiLineNext;
        // Pad unavailable samples with new value
        Int iNextOrTop = std::min<Int>(iNext, iLeftUnits);
        // fill left column
        while (iCurrJnit < iNextOrTop)
        {
          for (i=0; i<unitHeight; i++)
          {
            piAdiLineCur[i] = refSample;
          }
          piAdiLineCur += unitHeight;
          iCurrJnit++;
        }
        // fill top row
        while (iCurrJnit < iNext)
        {
          for (i=0; i<unitWidth; i++)
          {
            piAdiLineCur[i] = refSample;
          }
          piAdiLineCur += unitWidth;
          iCurrJnit++;
        }
      }
    }

    // pad all other reference samples.
    while (iCurrJnit < iTotalUnits)
    {
      if (!bNeighborFlags[iCurrJnit]) // samples not available
      {
        {
          const Int numSamplesInCurrUnit = (iCurrJnit >= iLeftUnits) ? unitWidth : unitHeight;
          const Pel refSample = *(piAdiLineCur-1);
          for (i=0; i<numSamplesInCurrUnit; i++)
          {
            piAdiLineCur[i] = refSample;
          }
          piAdiLineCur += numSamplesInCurrUnit;
          iCurrJnit++;
        }
      }
      else
      {
        piAdiLineCur += (iCurrJnit >= iLeftUnits) ? unitWidth : unitHeight;
        iCurrJnit++;
      }
    }

    // Copy processed samples

    piAdiLineTemp = piAdiLine + uiHeight + unitWidth - 2;
    // top left, top and top right samples
    for (i=0; i<uiWidth; i++)
    {
      piAdiTemp[i] = piAdiLineTemp[i];
    }

    piAdiLineTemp = piAdiLine + uiHeight - 1;
    for (i=1; i<uiHeight; i++)
    {
      piAdiTemp[i*uiWidth] = piAdiLineTemp[-i];
    }
  }
}

/** Get pointer to reference samples for intra prediction
 * \param uiDirMode   prediction mode index
 * \param log2BlkSize size of block (2 = 4x4, 3 = 8x8, 4 = 16x16, 5 = 32x32, 6 = 64x64)
 * \param piAdiBuf    pointer to unfiltered reference samples
 * \return            pointer to (possibly filtered) reference samples
 *
 * The prediction mode index is used to determine whether a smoothed reference sample buffer is returned.
 */

Bool TComPrediction::filteringIntraReferenceSamples(const ComponentID compID, UInt uiDirMode, UInt uiTuChWidth, UInt uiTuChHeight, const ChromaFormat chFmt, const Bool intraReferenceSmoothingDisabled)
{
  Bool bFilter;

  if (!filterIntraReferenceSamples(toChannelType(compID), chFmt, intraReferenceSmoothingDisabled))
  {
    bFilter=false;
  }
  else
  {
    assert(uiTuChWidth>=4 && uiTuChHeight>=4 && uiTuChWidth<128 && uiTuChHeight<128);

    if (uiDirMode == DC_IDX)
    {
      bFilter=false; //no smoothing for DC or LM chroma
    }
    else
    {
      Int diff = min<Int>(abs((Int) uiDirMode - HOR_IDX), abs((Int)uiDirMode - VER_IDX));
      UInt sizeIndex=g_aucConvertToBit[uiTuChWidth];
      assert(sizeIndex < MAX_INTRA_FILTER_DEPTHS);
      bFilter = diff > m_aucIntraFilter[toChannelType(compID)][sizeIndex];
    }
  }
  return bFilter;
}

Bool isAboveLeftAvailable( TComDataCU* pcCU, UInt uiPartIdxLT )
{
  Bool bAboveLeftFlag;
  UInt uiPartAboveLeft;
  TComDataCU* pcCUAboveLeft = pcCU->getPUAboveLeft( uiPartAboveLeft, uiPartIdxLT );
  if(pcCU->getSlice()->getPPS()->getConstrainedIntraPred())
  {
    bAboveLeftFlag = ( pcCUAboveLeft && pcCUAboveLeft->isConstrainedIntra( uiPartAboveLeft ) );
  }
  else
  {
    bAboveLeftFlag = (pcCUAboveLeft ? true : false);
  }
  return bAboveLeftFlag;
}

Int isAboveAvailable( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxRT, Bool *bValidFlags )
{
  const UInt uiRasterPartBegin = g_auiZscanToRaster[uiPartIdxLT];
  const UInt uiRasterPartEnd = g_auiZscanToRaster[uiPartIdxRT]+1;
  const UInt uiIdxStep = 1;
  Bool *pbValidFlags = bValidFlags;
  Int iNumIntra = 0;

  for ( UInt uiRasterPart = uiRasterPartBegin; uiRasterPart < uiRasterPartEnd; uiRasterPart += uiIdxStep )
  {
    UInt uiPartAbove;
    TComDataCU* pcCUAbove = pcCU->getPUAbove( uiPartAbove, g_auiRasterToZscan[uiRasterPart] );
    if(pcCU->getSlice()->getPPS()->getConstrainedIntraPred())
    {
      if ( pcCUAbove && pcCUAbove->isConstrainedIntra( uiPartAbove ) )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    else
    {
      if (pcCUAbove)
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    pbValidFlags++;
  }
  return iNumIntra;
}

Int isLeftAvailable( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxLB, Bool *bValidFlags )
{
  const UInt uiRasterPartBegin = g_auiZscanToRaster[uiPartIdxLT];
  const UInt uiRasterPartEnd = g_auiZscanToRaster[uiPartIdxLB]+1;
  const UInt uiIdxStep = pcCU->getPic()->getNumPartInWidth();
  Bool *pbValidFlags = bValidFlags;
  Int iNumIntra = 0;

  for ( UInt uiRasterPart = uiRasterPartBegin; uiRasterPart < uiRasterPartEnd; uiRasterPart += uiIdxStep )
  {
    UInt uiPartLeft;
    TComDataCU* pcCULeft = pcCU->getPULeft( uiPartLeft, g_auiRasterToZscan[uiRasterPart] );
    if(pcCU->getSlice()->getPPS()->getConstrainedIntraPred())
    {
      if ( pcCULeft && pcCULeft->isConstrainedIntra( uiPartLeft ) )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    else
    {
      if ( pcCULeft )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    pbValidFlags--; // opposite direction
  }

  return iNumIntra;
}

Int isAboveRightAvailable( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxRT, Bool *bValidFlags )
{
  const UInt uiNumUnitsInPU = g_auiZscanToRaster[uiPartIdxRT] - g_auiZscanToRaster[uiPartIdxLT] + 1;
  Bool *pbValidFlags = bValidFlags;
  Int iNumIntra = 0;

  for ( UInt uiOffset = 1; uiOffset <= uiNumUnitsInPU; uiOffset++ )
  {
    UInt uiPartAboveRight;
    TComDataCU* pcCUAboveRight = pcCU->getPUAboveRightAdi( uiPartAboveRight, uiPartIdxRT, uiOffset );
    if(pcCU->getSlice()->getPPS()->getConstrainedIntraPred())
    {
      if ( pcCUAboveRight && pcCUAboveRight->isConstrainedIntra( uiPartAboveRight ) )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    else
    {
      if ( pcCUAboveRight )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    pbValidFlags++;
  }

  return iNumIntra;
}

Int isBelowLeftAvailable( TComDataCU* pcCU, UInt uiPartIdxLT, UInt uiPartIdxLB, Bool *bValidFlags )
{
  const UInt uiNumUnitsInPU = (g_auiZscanToRaster[uiPartIdxLB] - g_auiZscanToRaster[uiPartIdxLT]) / pcCU->getPic()->getNumPartInWidth() + 1;
  Bool *pbValidFlags = bValidFlags;
  Int iNumIntra = 0;

  for ( UInt uiOffset = 1; uiOffset <= uiNumUnitsInPU; uiOffset++ )
  {
    UInt uiPartBelowLeft;
    TComDataCU* pcCUBelowLeft = pcCU->getPUBelowLeftAdi( uiPartBelowLeft, uiPartIdxLB, uiOffset );
    if(pcCU->getSlice()->getPPS()->getConstrainedIntraPred())
    {
      if ( pcCUBelowLeft && pcCUBelowLeft->isConstrainedIntra( uiPartBelowLeft ) )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    else
    {
      if ( pcCUBelowLeft )
      {
        iNumIntra++;
        *pbValidFlags = true;
      }
      else
      {
        *pbValidFlags = false;
      }
    }
    pbValidFlags--; // opposite direction
  }

  return iNumIntra;
}

#if PALETTE_MODE
Void TComPrediction::deriveRunLuma (TComDataCU* pcCU, Pel* pOrg,  Pel *pPalette,  Pel* pValue, Pel* pSPoint, Pel* pEscapeFlag, Pel* pPixelPredFlag,   Pel * paRecoValue, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize)
{
  UInt uiTotal = uiHeight * uiWidth, uiIdx = 0, uiStride = uiWidth;
  UInt uiStartPos = 0,  uiRun = 0, uiCopyRun = 0;
  Int iTemp = 0;

  //Test Run
  while (uiIdx < uiTotal)
  {
    uiStartPos = uiIdx;

    uiRun = 0;
    Bool RunValid = CalRunLuma     ( pcCU, pOrg, pPalette,   pValue, pSPoint, pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiRun);

    uiCopyRun = 0;
    Bool CopyValid = CalCopyLuma    ( pcCU, pOrg, pPalette,   pValue, pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiCopyRun);

    if (CopyValid == 0 && RunValid == 0)
    {
      pEscapeFlag[uiIdx] = 1;
      pPixelPredFlag[uiIdx] = CalcPixelPredLuma(pcCU, pOrg, pPalette, pValue, pEscapeFlag,   paRecoValue, uiWidth, uiHeight, uiStrideOrg, uiStartPos );
      pSPoint[uiIdx] = 0;
      uiIdx++;
    }
    else
    {
      pEscapeFlag[uiIdx] = 0;
      pPixelPredFlag[uiIdx] = 0;
      paRecoValue [uiIdx] = pPalette[pValue[uiIdx]];
      if ( CopyValid && (uiCopyRun + 2 > uiRun))
      {
        pSPoint[uiIdx] = 1;
        pRun[uiIdx]  = uiCopyRun-1;

        uiIdx++;
        iTemp = uiCopyRun - 1;
        while (iTemp>0)
        {
          pValue[uiIdx] = pValue[uiIdx - uiStride];
          paRecoValue[uiIdx] = pPalette[pValue[uiIdx]];
          pSPoint[uiIdx] = 1;
          pEscapeFlag[uiIdx] = 0;
          uiIdx++;
          iTemp--;
        }
      }
      else
      {
        pSPoint[uiIdx] = 0;
        pRun[uiIdx] = uiRun;
        uiIdx++;
        iTemp = uiRun;
        while (iTemp>0)
        {
          pValue[uiIdx] = pValue[uiIdx - 1];
          paRecoValue [uiIdx] = pPalette[pValue[uiIdx]];
          pSPoint[uiIdx] = 0;
          pEscapeFlag[uiIdx] = 0;
          uiIdx++;
          iTemp--;
        }
      }
    }
  }
  assert (uiIdx == uiTotal);
}

Void  TComPrediction:: deriveRunChroma (TComDataCU* pcCU, Pel* pOrg [2],  Pel *pPalette [2],  Pel* pValue[2], Pel* pSPoint, Pel* pEscapeFlag, Pel* pPixelPredFlag,   Pel * paRecoValue[2], Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize)
{

  UInt uiTotal = uiHeight * uiWidth, uiIdx = 0;
  UInt uiStartPos = 0,  uiRun = 0, uiCopyRun = 0;

  Int iTemp = 0;
  UInt uiStride = uiWidth;
  //Test Run
  while (uiIdx < uiTotal)
  {
    uiStartPos = uiIdx;
    uiRun = 0;

    Bool RunValid = false;
    RunValid = CalRunChroma     ( pcCU, pOrg, pPalette,   pValue[0], pSPoint, pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiRun);
 
    uiCopyRun = 0;
    Bool CopyValid = false;
    CopyValid =   CalCopyChroma   ( pcCU, pOrg, pPalette,   pValue[0], pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiCopyRun);
 
    if (CopyValid == 0 && RunValid == 0)
    {
      pEscapeFlag[uiIdx] = 1;
      pPixelPredFlag[uiIdx] = CalcPixelPredChroma(pcCU, pOrg, pPalette, pValue, pEscapeFlag,   paRecoValue, uiWidth, uiHeight, uiStrideOrg, uiStartPos );
      pSPoint[uiIdx] = 0;
      uiIdx++;
    }
    else
    {
      pEscapeFlag[uiIdx] = 0;
      pPixelPredFlag[uiIdx] = 0;
      paRecoValue[0] [uiIdx] = pPalette[0][pValue[0][uiIdx]];
      paRecoValue[1] [uiIdx] = pPalette[1][pValue[0][uiIdx]];

      if ( CopyValid && (uiCopyRun + 2 > uiRun))
      {
        pSPoint[uiIdx] = 1;
        pRun[uiIdx]  = uiCopyRun-1;
        uiIdx++;
        iTemp = uiCopyRun - 1;
        while (iTemp>0)
        {
          pValue[0][uiIdx] = pValue[0][uiIdx-uiStride];
          paRecoValue[0] [uiIdx] = pPalette[0][pValue[0][uiIdx]];
          paRecoValue[1] [uiIdx] = pPalette[1][pValue[0][uiIdx]];

          pSPoint[uiIdx] = 1;
          pEscapeFlag[uiIdx] = 0;
          uiIdx++;
          iTemp--;
        }
      }
      else
      {
        pSPoint[uiIdx] = 0;
        pRun[uiIdx]  = uiRun;
        uiIdx++;
        iTemp = uiRun;
        while (iTemp>0)
        {
          pValue[0][uiIdx] = pValue[0][uiIdx-1];
          paRecoValue[0] [uiIdx] = pPalette[0][pValue[0][uiIdx]];
          paRecoValue[1] [uiIdx] = pPalette[1][pValue[0][uiIdx]];

          pSPoint[uiIdx] = 0;
          pEscapeFlag[uiIdx] = 0;
          uiIdx++;
          iTemp--;
        }
      }
    }
  }

  assert (uiIdx == uiTotal);
}


Void  TComPrediction:: deriveRun (TComDataCU* pcCU, Pel* pOrg [3],  Pel *pPalette [3],  Pel* pValue, Pel* pSPoint, Pel* pEscapeFlag, Pel* pPixelPredFlag,  Pel** paPixelValue, Pel ** paRecoValue, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize)
{

  UInt   uiTotal = uiHeight * uiWidth, uiIdx = 0;
  UInt uiStartPos = 0,  uiRun = 0, uiCopyRun = 0;

  Int iTemp = 0;
#if !PLT_LAST_MERGE && !PLT_DERIVATION_MEAN
  UInt uiStride = uiWidth;
#endif
  //Test Run
  while (uiIdx < uiTotal)
  {
    uiStartPos = uiIdx;
    uiRun = 0;

    Bool RunValid = false;
    RunValid = CalRun     ( pcCU, pOrg, pPalette,   pValue, pSPoint, pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiRun);

    uiCopyRun = 0;
    Bool CopyValid = false;
    CopyValid =   CalCopy    ( pcCU, pOrg, pPalette,   pValue, pEscapeFlag, pRun,  uiWidth,  uiHeight, uiStrideOrg, uiPLTSize, uiStartPos, uiTotal, uiCopyRun);

    if (CopyValid == 0 && RunValid == 0)
    {
#if PLT_LAST_MERGE || PLT_DERIVATION_MEAN
			assert(pEscapeFlag[uiIdx] == 1);
#else
      pEscapeFlag[uiIdx] = 1;
#endif
      pPixelPredFlag[uiIdx] = CalcPixelPred(pcCU, pOrg, pPalette, pValue, pEscapeFlag,  paPixelValue, paRecoValue, uiWidth, uiHeight, uiStrideOrg, uiStartPos );

      pSPoint[uiIdx] = 0;
      uiIdx++;
    }
    else
    {
#if !PLT_LAST_MERGE && !PLT_DERIVATION_MEAN
      pEscapeFlag[uiIdx] = 0;
#endif
      pPixelPredFlag[uiIdx] = 0;

      if ( CopyValid && (uiCopyRun + 2 > uiRun))
      {
        pSPoint[uiIdx] = 1;
        pRun[uiIdx]  = uiCopyRun-1;
        uiIdx++;
        iTemp = uiCopyRun - 1;
        while (iTemp>0)
        {
#if !PLT_LAST_MERGE && !PLT_DERIVATION_MEAN
          pValue[uiIdx] = pValue[uiIdx-uiStride];
#endif
          pSPoint[uiIdx] = 1;
#if !PLT_LAST_MERGE && !PLT_DERIVATION_MEAN
          pEscapeFlag[uiIdx] = 0;
#endif
          uiIdx++;
          iTemp--;
        }
      }
      else
      {
        pSPoint[uiIdx] = 0;
        pRun[uiIdx] = uiRun;

        uiIdx++;
        iTemp = uiRun;
        while (iTemp > 0)
        {
#if !PLT_LAST_MERGE && !PLT_DERIVATION_MEAN
          pValue[uiIdx] = pValue[uiIdx - 1];
#endif
          pSPoint[uiIdx] = 0;
#if !PLT_LAST_MERGE && !PLT_DERIVATION_MEAN
          pEscapeFlag[uiIdx] = 0;
#endif
          uiIdx++;
          iTemp--;
        }
      }
    }

  }
  assert (uiIdx == uiTotal);
}

Void  TComPrediction::derivePLTLuma (TComDataCU* pcCU, Pel *Palette, Pel* pSrc,  UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{

  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless? 0: getPLTErrLimit();

  UInt uiTotalSize = uiHeight*uiWidth;
  SortingElementLuma *psList = new SortingElementLuma [uiTotalSize];
  SortingElementLuma sElement;

  UInt uiDictMaxSize = MAX_PLT_SIZE;
  SortingElementLuma *pListSort = new SortingElementLuma [MAX_PLT_SIZE+1];

  UInt uiIdx = 0;
  UInt uiShift = 0;
  for( UInt uiY = 0; uiY < uiHeight; uiY++ )
  {
    uiShift     =  uiY * uiWidth;
    for( UInt uiX = 0; uiX < uiWidth; uiX++ )
    {
      sElement.setAll (pSrc[uiShift + uiX]);
      Int i = 0;
      for ( i = uiIdx-1; i>=0; i--)
      { 
        if (psList[i].almostEqualData(sElement, iErrorLimit))
        {

          psList[i].uiCnt++;

          break;
        }
      }

      if (i == -1)
      {
        psList[uiIdx].copyDataFrom(sElement);

        psList[uiIdx].uiCnt++;

        uiIdx ++;
      }
    }
  }


  for (Int i = 0; i < uiDictMaxSize; i++)
  {
    pListSort[i].uiCnt  = 0;
    pListSort[i].setAll(0) ;
  }

  for (Int i = 0; i < uiTotalSize; i++)
  {
    pListSort[uiDictMaxSize].copyAllFrom (psList[i]);

    for (Int j = uiDictMaxSize; j > 0; j--) 
    {
      if (pListSort[j].uiCnt > pListSort[j-1].uiCnt)
      {
        sElement.copyAllFrom (pListSort[j-1]) ;
        pListSort[j-1].copyAllFrom (pListSort[j]) ;
        pListSort[j].copyAllFrom (sElement) ;
      }
      else
      {
        break;
      }
    }
  }

  uiPLTSize = MAX_PLT_SIZE;
  for (Int i = 0; i < MAX_PLT_SIZE; i++)
  {
    Palette[i] = pListSort[i].uiDataLuma;
    if (pListSort[i].uiCnt == 0  )
    {
      uiPLTSize = i;
      break;
    }
  }

  delete [] psList;
  delete [] pListSort;


}

Void  TComPrediction::derivePLTChroma (TComDataCU* pcCU, Pel *Palette[2], Pel* pSrc[2],  UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{

  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless? 0: getPLTErrLimit();

  UInt uiTotalSize = uiHeight*uiWidth;
  SortingElementChroma *psList = new SortingElementChroma [uiTotalSize];
  SortingElementChroma sElement;

  UInt uiDictMaxSize = MAX_PLT_SIZE;
  SortingElementChroma *pListSort = new SortingElementChroma [MAX_PLT_SIZE+1];

  UInt uiIdx = 0;
  //UInt uiShift = 0, uiCbCrShift = 0;
  for( UInt uiY = 0; uiY < uiHeight; uiY++ )
  {
    UInt uiShift     =  uiY * uiWidth;
    for( UInt uiX = 0; uiX < uiWidth; uiX++ )
    {
      sElement.setAll (pSrc[0][uiShift + uiX], pSrc[1][uiShift + uiX]);
      Int i = 0;
      for ( i = uiIdx-1; i>=0; i--)
      { 
        if (psList[i].almostEqualData(sElement, iErrorLimit))
        {

          psList[i].uiCnt++;

          break;
        }
      }

      if (i == -1)
      {
        psList[uiIdx].copyDataFrom(sElement);

        psList[uiIdx].uiCnt++;

        uiIdx ++;

      }
    }
  }


  for (Int i = 0; i < uiDictMaxSize; i++)
  {
    pListSort[i].uiCnt  = 0;
    pListSort[i].setAll(0, 0) ;
  }

  for (Int i = 0; i < uiTotalSize; i++)
  {
    pListSort[uiDictMaxSize].copyAllFrom (psList[i]);

    for (Int j = uiDictMaxSize; j > 0; j--) 
    {
      if (pListSort[j].uiCnt > pListSort[j-1].uiCnt)
      {
        sElement.copyAllFrom (pListSort[j-1]) ;
        pListSort[j-1].copyAllFrom (pListSort[j]) ;
        pListSort[j].copyAllFrom (sElement) ;
      }
      else
      {
        break;
      }
    }
  }

  uiPLTSize = MAX_PLT_SIZE;
  for (Int i = 0; i < MAX_PLT_SIZE; i++)
  {
    Palette[0][i] = pListSort[i].uiData[0];
    Palette[1][i] = pListSort[i].uiData[1];
    if (pListSort[i].uiCnt == 0  )
    {
      uiPLTSize = i;
      break;
    }
  }

  delete [] psList;
  delete [] pListSort;


}



Void   TComPrediction:: reorderPLTLuma (TComDataCU* pcCU, Pel *pPalette)
{
  UInt uiPLTSizePrev, uiDictMaxSize;
  Pel * pPalettePrev;
  Pel pPaletteTemp[MAX_PLT_SIZE];


  pPalettePrev = pcCU->getPLTPred (pcCU, pcCU->getZorderIdxInCU(), COMPONENT_Y, uiPLTSizePrev);
  for (UInt i = 0; i < MAX_PLT_SIZE; i++)
  {
    pPaletteTemp[i] = pPalette[i];
  }


  pcCU->getPLTPred (pcCU, pcCU->getZorderIdxInCU(), COMPONENT_Y, uiPLTSizePrev);
  uiDictMaxSize = pcCU->getPLTSize(COMPONENT_Y, 0);

  UInt uiIdxPrev = 0, uiIdxCurr = 0;
  Bool bReused = false;
  Bool bPredicted [MAX_PLT_SIZE+1], bReusedPrev [MAX_PLT_SIZE+1];

  memset (bReusedPrev, 0, MAX_PLT_SIZE+1);
  memset (bPredicted,  0, MAX_PLT_SIZE+1);


  UInt uiNumPLTRceived = uiDictMaxSize , uiNumPLTPredicted = 0;

  for (uiIdxPrev = 0; uiIdxPrev < uiPLTSizePrev; uiIdxPrev ++)
  {
    bReused = false;
    //Int iCounter = 0;
    for (uiIdxCurr = 0; uiIdxCurr < uiDictMaxSize; uiIdxCurr ++)
    {
      if (pPalettePrev[uiIdxPrev] == pPalette[uiIdxCurr] )
      {
        bReused = true;
        break;
      }

    }
    bReusedPrev[uiIdxPrev] = bReused;
    bPredicted [uiIdxCurr] = bReused;
    if (bPredicted[uiIdxCurr] )
    {
      uiNumPLTRceived--;
      uiNumPLTPredicted ++;
    }
  }
  for (uiIdxPrev = 0; uiIdxPrev < MAX_PLT_SIZE; uiIdxPrev ++)
  {

    pcCU->setPrevPLTReusedFlagSubParts   ( COMPONENT_Y,  bReusedPrev[uiIdxPrev], uiIdxPrev, 0,  pcCU->getDepth(0)  );
  }
  uiIdxCurr = 0;
  for (UInt uiPrevIdx = 0; uiPrevIdx < uiPLTSizePrev; uiPrevIdx ++)
  {
    if (bReusedPrev[uiPrevIdx])
    {

      pPalette[uiIdxCurr] = pPalettePrev[uiPrevIdx];
      uiIdxCurr ++;
    }
  }

  //UInt uiReceivedIdx = 0;
  for (UInt uiIdx = 0; uiIdx < uiDictMaxSize; uiIdx ++)
  {
    if (bPredicted[uiIdx] == 0)
    {
      pPalette[uiIdxCurr] = pPaletteTemp[uiIdx];
      uiIdxCurr++;
    }

  }
}
Void   TComPrediction:: reorderPLTChroma (TComDataCU* pcCU, Pel *pPalette[2])
{

  UInt uiPLTSizePrev, uiDictMaxSize;
  Pel * pPalettePrev[2];
  Pel pPaletteTemp[2][MAX_PLT_SIZE];

    for (UInt comp=0; comp<MAX_NUM_COMPONENT-1; comp++)
    {
      pPalettePrev[comp] = pcCU->getPLTPred (pcCU,  pcCU->getZorderIdxInCU(), comp+1, uiPLTSizePrev);
      for (UInt i = 0; i < MAX_PLT_SIZE; i++)
      {
        pPaletteTemp[comp][i] = pPalette[comp][i];
      }
    }
 
  pcCU->getPLTPred (pcCU, pcCU->getZorderIdxInCU(), COMPONENT_Cb, uiPLTSizePrev);
  uiDictMaxSize = pcCU->getPLTSize(COMPONENT_Cb, 0);

  UInt uiIdxPrev = 0, uiIdxCurr = 0;
  Bool bReused = false;
  Bool bPredicted [MAX_PLT_SIZE+1], bReusedPrev [MAX_PLT_SIZE+1];

  memset (bReusedPrev, 0, MAX_PLT_SIZE+1);
  memset (bPredicted,  0, MAX_PLT_SIZE+1);


  UInt uiNumPLTRceived = uiDictMaxSize , uiNumPLTPredicted = 0;

  for (uiIdxPrev = 0; uiIdxPrev < uiPLTSizePrev; uiIdxPrev ++)
  {
    bReused = false;
    Int iCounter = 0;
    for (uiIdxCurr = 0; uiIdxCurr < uiDictMaxSize; uiIdxCurr ++)
    {
      iCounter = 0;

      for (UInt comp=0; comp<MAX_NUM_COMPONENT-1; comp++)
      {
        if (pPalettePrev[comp][uiIdxPrev] == pPalette[comp][uiIdxCurr] )
        {
          iCounter ++;
        }
      }
      if (iCounter == 2)
      {
        bReused = true;
        break;
      }
    }
    bReusedPrev[uiIdxPrev] = bReused;
    bPredicted [uiIdxCurr] = bReused;
    if (bPredicted[uiIdxCurr] )
    {
      uiNumPLTRceived--;
      uiNumPLTPredicted ++;
    }
  }
  for (uiIdxPrev = 0; uiIdxPrev < MAX_PLT_SIZE; uiIdxPrev ++)
  {
    for (UInt comp=0; comp<MAX_NUM_COMPONENT-1; comp++)
    {
      pcCU->setPrevPLTReusedFlagSubParts   ( comp+1,  bReusedPrev[uiIdxPrev], uiIdxPrev, 0,  pcCU->getDepth(0)  );
    }
  }
  uiIdxCurr = 0;
  for (UInt uiPrevIdx = 0; uiPrevIdx < uiPLTSizePrev; uiPrevIdx ++)
  {
    if (bReusedPrev[uiPrevIdx])
    {
      for (UInt comp=0; comp<MAX_NUM_COMPONENT-1; comp++)
      {
        pPalette[comp][uiIdxCurr] = pPalettePrev[comp][uiPrevIdx];
      }
      uiIdxCurr ++;
    }
  }

  //UInt uiReceivedIdx = 0;
  for (UInt uiIdx = 0; uiIdx < uiDictMaxSize; uiIdx ++)
  {
    if (bPredicted[uiIdx] == 0)
    {
      for (UInt comp=0; comp<MAX_NUM_COMPONENT-1; comp++)
      {
        pPalette[comp][uiIdxCurr] = pPaletteTemp[comp][uiIdx];
      }
      uiIdxCurr++;
    }

  }
}

Void  TComPrediction::reorderPLT (TComDataCU* pcCU, Pel *pPalette[3])
{

    UInt uiPLTSizePrev, uiDictMaxSize;
    Pel * pPalettePrev[3];
    Pel pPaletteTemp[3][MAX_PLT_SIZE];


    for (UInt comp=0; comp<MAX_NUM_COMPONENT; comp++)
    {
      pPalettePrev[comp] = pcCU->getPLTPred (pcCU,  pcCU->getZorderIdxInCU(), comp, uiPLTSizePrev);
      for (UInt i = 0; i < MAX_PLT_SIZE; i++)
      {
        pPaletteTemp[comp][i] = pPalette[comp][i];
      }
    }
 
  pcCU->getPLTPred (pcCU, pcCU->getZorderIdxInCU(), COMPONENT_Y, uiPLTSizePrev);
  uiDictMaxSize = pcCU->getPLTSize(COMPONENT_Y, 0);

  UInt uiIdxPrev = 0, uiIdxCurr = 0;
  Bool bReused = false;
#if PLT_LAST_MERGE
	Bool bPredicted [MAX_PLT_SIZE+1], bReusedPrev [MAX_PLT_DICT_SIZE+1];
	memset (bReusedPrev, 0, (uiPLTSizePrev+1)*sizeof(Bool));
	memset (bPredicted,  0, sizeof(bPredicted));
#else
  Bool bPredicted [MAX_PLT_SIZE+1], bReusedPrev [MAX_PLT_SIZE+1];
  memset (bReusedPrev, 0, MAX_PLT_SIZE+1);
  memset (bPredicted,  0, MAX_PLT_SIZE+1);
#endif

  UInt uiNumPLTRceived = uiDictMaxSize , uiNumPLTPredicted = 0;
#if PLT_DERIVATION_MEAN || PLT_LAST_MERGE
	for (uiIdxCurr = 0; uiIdxCurr < uiDictMaxSize; uiIdxCurr ++)
	{
		bReused = false;
		for (uiIdxPrev = 0; uiIdxPrev < uiPLTSizePrev; uiIdxPrev ++)
		{
			UInt comp;
			for (comp=0; comp<MAX_NUM_COMPONENT; comp++)
			{
				if (pPalettePrev[comp][uiIdxPrev] != pPalette[comp][uiIdxCurr] )
					break;
			}
			if (comp == MAX_NUM_COMPONENT)
			{
				bReused = true;
				break;
			}
		}
		bReusedPrev[uiIdxPrev] = bReused;
		bPredicted [uiIdxCurr] = bReused;
		if (bPredicted[uiIdxCurr] )
		{
			uiNumPLTRceived--;
			uiNumPLTPredicted ++;
		}
	}
#else
  for (uiIdxPrev = 0; uiIdxPrev < uiPLTSizePrev; uiIdxPrev ++)
  {
    bReused = false;
    Int iCounter = 0;
    for (uiIdxCurr = 0; uiIdxCurr < uiDictMaxSize; uiIdxCurr ++)
    {
      iCounter = 0;

      for (UInt comp=0; comp<MAX_NUM_COMPONENT; comp++)
      {
        if (pPalettePrev[comp][uiIdxPrev] == pPalette[comp][uiIdxCurr] )
        {
          iCounter ++;
        }
      }
      if (iCounter == 3)
      {
        bReused = true;
        break;
      }
    }
    bReusedPrev[uiIdxPrev] = bReused;
    bPredicted [uiIdxCurr] = bReused;
    if (bPredicted[uiIdxCurr] )
    {
      uiNumPLTRceived--;
      uiNumPLTPredicted ++;
    }
  }
#endif
#if PLT_LAST_MERGE
	for (uiIdxPrev = 0; uiIdxPrev < pcCU->getSlice()->getPPS()->getPLTDictionarySize(); uiIdxPrev ++)
#else
  for (uiIdxPrev = 0; uiIdxPrev < MAX_PLT_SIZE; uiIdxPrev ++)
#endif
  {
    for (UInt comp=0; comp<MAX_NUM_COMPONENT; comp++)
    {
      pcCU->setPrevPLTReusedFlagSubParts   ( comp,  bReusedPrev[uiIdxPrev], uiIdxPrev, 0,  pcCU->getDepth(0)  );
    }
  }
  uiIdxCurr = 0;
  for (UInt uiPrevIdx = 0; uiPrevIdx < uiPLTSizePrev; uiPrevIdx ++)
  {
    if (bReusedPrev[uiPrevIdx])
    {
      for (UInt comp=0; comp<MAX_NUM_COMPONENT; comp++)
      {
        pPalette[comp][uiIdxCurr] = pPalettePrev[comp][uiPrevIdx];
      }
      uiIdxCurr ++;
    }
  }

  //UInt uiReceivedIdx = 0;
  for (UInt uiIdx = 0; uiIdx < uiDictMaxSize; uiIdx ++)
  {
    if (bPredicted[uiIdx] == 0)
    {
      for (UInt comp=0; comp<MAX_NUM_COMPONENT; comp++)
      {
        pPalette[comp][uiIdxCurr] = pPaletteTemp[comp][uiIdx];
      }
      uiIdxCurr++;
    }

  }
}

#if !PLT_DERIVATION_MEAN
Void  TComPrediction::derivePLT (TComDataCU* pcCU, Pel *Palette[3], Pel* pSrc[3],  UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{

  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless? 0: getPLTErrLimit();

  UInt uiTotalSize = uiHeight*uiWidth;
  SortingElement *psList = new SortingElement [uiTotalSize];
  SortingElement sElement;
  UInt uiDictMaxSize = MAX_PLT_SIZE;
  SortingElement *pListSort = new SortingElement [MAX_PLT_SIZE+1];
  UInt uiIdx = 0;
  UInt uiPos;
  for( UInt uiY = 0; uiY < uiHeight; uiY++ )
  {
    for( UInt uiX = 0; uiX < uiWidth; uiX++ )
    {
      uiPos = uiY * uiWidth+ uiX;
      sElement.setAll (pSrc[0][uiPos], pSrc[1][uiPos], pSrc[2][uiPos]);
      Int i = 0;
      for ( i = uiIdx-1; i>=0; i--)
      { 
        if (psList[i].almostEqualData(sElement, iErrorLimit))
        {

          psList[i].uiCnt++;

          break;
        }
      }
      if (i == -1)
      {
        psList[uiIdx].copyDataFrom(sElement);

        psList[uiIdx].uiCnt++;

        uiIdx ++;

      }
    }
  }

  for (Int i = 0; i < uiDictMaxSize; i++)
  {
    pListSort[i].uiCnt  = 0;
    pListSort[i].setAll(0, 0, 0) ;
  }

  //bubble sorting
  for (Int i = 0; i < uiTotalSize; i++)
  {
    pListSort[uiDictMaxSize].copyAllFrom (psList[i]);
    for (Int j = uiDictMaxSize; j > 0; j--) 
    {
      if (pListSort[j].uiCnt > pListSort[j-1].uiCnt)
      {
        sElement.copyAllFrom (pListSort[j-1]) ;
        pListSort[j-1].copyAllFrom (pListSort[j]) ;
        pListSort[j].copyAllFrom (sElement) ;
      }
      else
      {
        break;
      }
    }
  }

  uiPLTSize = MAX_PLT_SIZE;
  for (Int i = 0; i < MAX_PLT_SIZE; i++)
  {
    Palette[0][i] = pListSort[i].uiData[0];
    Palette[1][i] = pListSort[i].uiData[1];
    Palette[2][i] = pListSort[i].uiData[2];
    if (pListSort[i].uiCnt == 0  )
    {
      uiPLTSize = i;
      break;
    }
  }

  delete [] psList;
  delete [] pListSort;


}
#endif
Bool  TComPrediction::CalcPixelPredLuma(TComDataCU* pcCU, Pel* pOrg, Pel *pPalette, Pel* pValue, Pel * pEscapeFlag,   Pel*paRecoValue, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiIdx )
{
  Bool bLossless = pcCU->getCUTransquantBypass (0);
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
  const QpParam cQP(*pcCU, ComponentID(0));
  Int iQP = cQP.Qp;
#else
  Int iQP = pcCU->getQP(0) ;
#endif
  Int iQPrem = iQP%6;
  Int iQPper = iQP/6;
  Int quantiserScale =  g_quantScales[iQPrem];
  const Int quantiserRightShift = QUANT_SHIFT + iQPper;
  const Int rightShiftOffset = 1<< (quantiserRightShift-1);
  const Int InvquantiserRightShift = (IQUANT_SHIFT - iQPper);
  Int iAdd =  InvquantiserRightShift == 0? 0: 1 << (InvquantiserRightShift - 1);

#if !PLT_ESCAPE_NO_PREDICTION
  UInt uiSADNoPred;
  UInt uiSADPred;
  UInt uiPredTop;
  UInt uiStride = uiWidth;
  Int iPredError;
#endif  
  UInt uiY, uiX;
  uiY = uiIdx/uiWidth;
  uiX = uiIdx%uiWidth;

  UInt uiYIdxRaster = uiY*uiStrideOrg + uiX;
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
  UInt uiMaxBit[3];
  pcCU->xCalcMaxBits(pcCU, uiMaxBit);
#endif

#if PLT_ESCAPE_NO_PREDICTION
  if (bLossless)
  {
    paRecoValue[uiYIdxRaster] = pValue[uiYIdxRaster] = pOrg[uiYIdxRaster];
  }
  else
  {
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
    pValue[uiYIdxRaster] = Pel(ClipBD<Int>((pOrg[uiYIdxRaster] * quantiserScale + rightShiftOffset) >> quantiserRightShift, uiMaxBit[0]));
#else
    pValue[uiYIdxRaster] = (pOrg[uiYIdxRaster] * quantiserScale + rightShiftOffset) >> quantiserRightShift;
#endif

    paRecoValue[uiYIdxRaster] = (pValue[uiYIdxRaster] * g_invQuantScales[iQPrem] + iAdd) >> InvquantiserRightShift;
    paRecoValue[uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[uiYIdxRaster], g_bitDepth[0]));
  }

  return false; //do not use pixel prediction

#else

  if (uiY > 0)
  {
    uiSADNoPred = pOrg [uiYIdxRaster];
    uiPredTop = pEscapeFlag[uiIdx-uiStride]? paRecoValue [uiYIdxRaster - uiStride]: pPalette[pValue[uiIdx-uiStride]];
    iPredError = pOrg[uiYIdxRaster] - uiPredTop;
    uiSADPred  = iPredError>0? iPredError: -iPredError;
  }
  else
  {
    uiSADNoPred = 0;
    uiSADPred = 100;
  }
  if (uiSADNoPred< uiSADPred )
  {
    if (bLossless)
    {
      paRecoValue[uiYIdxRaster] = pValue [uiYIdxRaster] =  pOrg[uiYIdxRaster];
    }
    else
    {
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
      pValue[uiYIdxRaster] = Pel(ClipBD<Int>((pOrg[uiYIdxRaster] * quantiserScale + rightShiftOffset) >> quantiserRightShift, uiMaxBit[0]));
#else
      pValue[uiYIdxRaster] =  (pOrg[uiYIdxRaster]   *quantiserScale + rightShiftOffset)>>quantiserRightShift;
#endif
      paRecoValue[uiYIdxRaster] = (pValue[uiYIdxRaster]*g_invQuantScales[iQPrem] + iAdd)>>InvquantiserRightShift;
      paRecoValue[uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[uiYIdxRaster], g_bitDepth[0]));
    }
    return false;
  }
  else
  {

    if (bLossless)
    {
      pValue [uiYIdxRaster] =  pOrg[uiYIdxRaster]    - uiPredTop;
      paRecoValue[uiYIdxRaster] = pOrg[uiYIdxRaster];
    }
    else
    {
      Int iError = pOrg[uiYIdxRaster]    - uiPredTop;
      Int iSign = iError<0? -1: 1;
      UInt uiLevel = iSign*iError;
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
      UInt uiQLevel = Pel(ClipBD<Int>((uiLevel*quantiserScale + rightShiftOffset) >> quantiserRightShift, uiMaxBit[0]));
#else
      UInt uiQLevel = (uiLevel   *quantiserScale + rightShiftOffset)>>quantiserRightShift;
#endif
      pValue[uiYIdxRaster] = iSign * uiQLevel;
      paRecoValue[uiYIdxRaster] =  iSign * ((uiQLevel*g_invQuantScales[iQPrem] + iAdd)>>InvquantiserRightShift);
      paRecoValue[uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[uiYIdxRaster] + uiPredTop, g_bitDepth[0]));
    }

    return true;
  }
#endif
}
Bool  TComPrediction::CalcPixelPredChroma(TComDataCU* pcCU, Pel* pOrg [2], Pel *pPalette[2], Pel* pValue[2], Pel * pEscapeFlag,   Pel*paRecoValue[2], UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiIdx )
{

  Bool bLossless = pcCU->getCUTransquantBypass (0);
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
  const QpParam cQP(*pcCU, ComponentID(1));//chroma
  Int iQP = cQP.Qp;
#else
  Int iQP = pcCU->getQP(0) ;
#endif
  Int iQPrem = iQP%6;
  Int iQPper = iQP/6;
  Int quantiserScale =  g_quantScales[iQPrem];
  const Int quantiserRightShift = QUANT_SHIFT + iQPper;
  const Int rightShiftOffset = 1<< (quantiserRightShift-1);
  const Int InvquantiserRightShift = (IQUANT_SHIFT - iQPper);
  Int iAdd =  InvquantiserRightShift == 0? 0: 1 << (InvquantiserRightShift - 1);
#if !PLT_ESCAPE_NO_PREDICTION
  UInt uiSADNoPred;
  UInt uiSADPred;
  UInt uiPredTop[2];
  UInt uiStride = uiWidth ;
  Int iPredError;
#endif
  UInt uiY, uiX;
  uiY = uiIdx/uiWidth;
  uiX = uiIdx%uiWidth;

  UInt uiYIdxRaster    = uiY*uiStrideOrg + uiX;

#if PLT_ESCAPE_ADAPTIVE_BD_FLC
  UInt uiMaxBit[3];
  pcCU->xCalcMaxBits(pcCU, uiMaxBit);
#endif

#if PLT_ESCAPE_NO_PREDICTION
  if (bLossless)
  {
    for (int i = 0; i < 2; i++)
    {
      pValue[i][uiYIdxRaster] = pOrg[i][uiYIdxRaster];
      paRecoValue[i][uiYIdxRaster] = pOrg[i][uiYIdxRaster];
    }
  }
  else
  {
    for (int i = 0; i < 2; i++)
    {
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
      pValue[i][uiYIdxRaster] = Pel(ClipBD<Int>((pOrg[i][uiYIdxRaster] * quantiserScale + rightShiftOffset) >> quantiserRightShift, uiMaxBit[1]));
#else
      pValue[i][uiYIdxRaster] = (pOrg[i][uiYIdxRaster] * quantiserScale + rightShiftOffset) >> quantiserRightShift;
#endif

      paRecoValue[i][uiYIdxRaster] = (pValue[i][uiYIdxRaster] * g_invQuantScales[iQPrem] + iAdd) >> InvquantiserRightShift;
      paRecoValue[i][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[i][uiYIdxRaster], g_bitDepth[1]));
    }
  }

  return false; //do not use pixel prediction

#else

  if (uiY > 0)
  {
    uiSADNoPred = uiSADPred =0;
    for (int i = 0; i < 2; i++)
    {
      uiSADNoPred += pOrg[i] [uiYIdxRaster];
      uiPredTop[i] = pEscapeFlag[uiIdx-uiStride]? paRecoValue[i] [uiYIdxRaster - uiStride]: pPalette[i][pValue[0][uiIdx-uiStride]];
      iPredError = pOrg[i] [uiYIdxRaster] - uiPredTop[i];
      uiSADPred  += iPredError>0? iPredError: -iPredError;
    }
  }
  else
  {
    uiSADNoPred = 0;
    uiSADPred = 100;
  }
  if (uiSADNoPred< uiSADPred )
  {
    if (bLossless)
    {
      for (int i = 0; i < 2; i++)
      {
        pValue[i] [uiYIdxRaster] =  pOrg[i][uiYIdxRaster];
        paRecoValue[i][uiYIdxRaster] = pOrg[i][uiYIdxRaster];
      }
    }
    else
    {
      for (int i = 0; i < 2; i++)
      {
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
        pValue[i][uiYIdxRaster] = Pel(ClipBD<Int>((pOrg[i][uiYIdxRaster] * quantiserScale + rightShiftOffset) >> quantiserRightShift, uiMaxBit[1]));
#else
        pValue[i] [uiYIdxRaster] =  (pOrg[i][uiYIdxRaster]   *quantiserScale + rightShiftOffset)>>quantiserRightShift;
#endif
        paRecoValue[i][uiYIdxRaster] = (pValue[i] [uiYIdxRaster]*g_invQuantScales[iQPrem] + iAdd)>>InvquantiserRightShift;
        paRecoValue[i][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[i][uiYIdxRaster], g_bitDepth[1]));
      }
    }

    return false;
  }
  else
  {

    if (bLossless)
    {
      for (int i = 0; i < 2; i++)
      {
        pValue[i] [uiYIdxRaster] =  pOrg[i][uiYIdxRaster] - uiPredTop[i];
        paRecoValue[i][uiYIdxRaster] = pOrg[i][uiYIdxRaster];
      }
    }
    else
    {
      Int iError, iSign;
      UInt uiLevel, uiQLevel;

      for (int i = 0; i < 2; i++)
      {
        iError   = pOrg[i][uiYIdxRaster]    - uiPredTop[i];
        iSign    = iError<0? -1: 1;
        uiLevel  = iSign*iError;
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
        uiQLevel = Pel(ClipBD<Int>((uiLevel   *quantiserScale + rightShiftOffset) >> quantiserRightShift, uiMaxBit[1]));
#else
        uiQLevel = (uiLevel   *quantiserScale + rightShiftOffset)>>quantiserRightShift;
#endif
        pValue[i] [uiYIdxRaster] = iSign * uiQLevel;
        paRecoValue[i][uiYIdxRaster] =  iSign * ((uiQLevel*g_invQuantScales[iQPrem] + iAdd)>>InvquantiserRightShift);
        paRecoValue[i][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[i][uiYIdxRaster] + uiPredTop[i], g_bitDepth[1]));
      }
    }

    return true;
  }
#endif
}


Bool TComPrediction:: CalcPixelPred(TComDataCU* pcCU, Pel* pOrg [3], Pel *pPalette[3], Pel* pValue, Pel * pEscapeFlag,  Pel*paPixelValue[3], Pel * paRecoValue[3], UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiIdx )
{
  Bool bLossless = pcCU->getCUTransquantBypass (0);
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
  Int iQP[3];
  Int iQPrem[3];
  Int iQPper[3];
  Int quantiserScale[3];
  Int quantiserRightShift[3];
  Int rightShiftOffset[3];
  Int InvquantiserRightShift[3];
  Int iAdd[3];
  for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch++)
  {
    QpParam cQP(*pcCU, ComponentID(ch));
    iQP[ch] = cQP.Qp;
    iQPrem[ch] = iQP[ch] % 6;
    iQPper[ch] = iQP[ch] / 6;
    quantiserScale[ch] = g_quantScales[iQPrem[ch]];
    quantiserRightShift[ch] = QUANT_SHIFT + iQPper[ch];
    rightShiftOffset[ch] = 1 << (quantiserRightShift[ch] - 1);
    InvquantiserRightShift[ch] = (IQUANT_SHIFT - iQPper[ch]);
    iAdd[ch] = InvquantiserRightShift[ch] == 0 ? 0 : 1 << (InvquantiserRightShift[ch] - 1);
  }
#else
  Int iQP = pcCU->getQP(0) ;
  Int iQPrem = iQP % 6;
  Int iQPper = iQP / 6;
  Int quantiserScale = g_quantScales[iQPrem];
  const Int quantiserRightShift = QUANT_SHIFT + iQPper;
  const Int rightShiftOffset = 1 << (quantiserRightShift-1);
  const Int InvquantiserRightShift = (IQUANT_SHIFT - iQPper);
  Int iAdd = InvquantiserRightShift == 0? 0: 1 << (InvquantiserRightShift - 1);
#endif
#if !PLT_ESCAPE_NO_PREDICTION
  UInt uiSADNoPred;
  UInt uiSADPred;
  UInt uiPredTop[3];
  UInt uiStride = uiWidth ;
  Int iPredError;
#endif
  UInt uiY, uiX;
  uiY = uiIdx / uiWidth;
  uiX = uiIdx % uiWidth;
  UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;

#if PLT_ESCAPE_ADAPTIVE_BD_FLC
  UInt uiMaxBit[3];
  pcCU->xCalcMaxBits(pcCU, uiMaxBit);
#endif

#if PLT_ESCAPE_NO_PREDICTION
  if (bLossless)
  {
    for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
    {
      paPixelValue[ch][uiYIdxRaster] =  pOrg[ch][uiYIdxRaster];
      paRecoValue[ch][uiYIdxRaster] = pOrg[ch][uiYIdxRaster];
    }
  }
  else
  {
    for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
    {
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
      paPixelValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>((pOrg[ch][uiYIdxRaster] * quantiserScale[ch] + rightShiftOffset[ch]) >> quantiserRightShift[ch], uiMaxBit[ch]));
      paRecoValue[ch][uiYIdxRaster] = (paPixelValue[ch] [uiYIdxRaster]*g_invQuantScales[iQPrem[ch]] + iAdd[ch])>>InvquantiserRightShift[ch];
      paRecoValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[ch][uiYIdxRaster], g_bitDepth[ch? 1:0]));
#else
      paPixelValue[ch] [uiYIdxRaster] = (pOrg[ch][uiYIdxRaster] * quantiserScale + rightShiftOffset)>>quantiserRightShift;
      paRecoValue[ch][uiYIdxRaster] = (paPixelValue[ch] [uiYIdxRaster]*g_invQuantScales[iQPrem] + iAdd)>>InvquantiserRightShift;
      paRecoValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[ch][uiYIdxRaster], g_bitDepth[ch? 1:0]));
#endif
    }
  }

  return false; //do not use pixel prediction
#else
  if (uiY == 0)
  {
    if (bLossless)
    {
      for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
      {
        paPixelValue[ch] [uiYIdxRaster] =  pOrg[ch][uiYIdxRaster];
        paRecoValue[ch][uiYIdxRaster]   = pOrg[ch][uiYIdxRaster];
      }
    }
    else
    {
      for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
      {
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
        paPixelValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>((pOrg[ch][uiYIdxRaster] * quantiserScale[ch] + rightShiftOffset[ch]) >> quantiserRightShift[ch], uiMaxBit[ch]));
        paRecoValue[ch][uiYIdxRaster] = (paPixelValue[ch][uiYIdxRaster]*g_invQuantScales[iQPrem[ch]] + iAdd[ch])>>InvquantiserRightShift[ch];
        paRecoValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[ch][uiYIdxRaster], g_bitDepth[ch? 1:0]));
#else
        paPixelValue[ch] [uiYIdxRaster] =  (pOrg[ch][uiYIdxRaster]   *quantiserScale + rightShiftOffset)>>quantiserRightShift;
        paRecoValue[ch][uiYIdxRaster] = (paPixelValue[ch] [uiYIdxRaster]*g_invQuantScales[iQPrem] + iAdd)>>InvquantiserRightShift;
        paRecoValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[ch][uiYIdxRaster], g_bitDepth[ch? 1:0]));
#endif
      }
    }
    return false;
  }
  uiSADNoPred = uiSADPred = 0;

  for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
  {
    uiSADNoPred += pOrg[ch] [uiYIdxRaster];
    uiPredTop[ch] = (pEscapeFlag[uiIdx-uiStride]? paRecoValue[ch] [uiYIdxRaster - uiStride]: pPalette[ch][pValue[uiIdx-uiStride]]);
    iPredError = pOrg[ch] [uiYIdxRaster] - uiPredTop[ch];
    uiSADPred  += (iPredError>0? iPredError: -iPredError);
  }

  if (uiSADNoPred< uiSADPred)
  {
    if (bLossless)
    {
      for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
      {
      paPixelValue[ch] [uiYIdxRaster] =  pOrg[ch][uiYIdxRaster];
      paRecoValue[ch][uiYIdxRaster] = pOrg[ch][uiYIdxRaster];
      }
    }
    else
    {
      for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
      {
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
        paPixelValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>((pOrg[ch][uiYIdxRaster] * quantiserScale[ch] + rightShiftOffset[ch]) >> quantiserRightShift[ch], uiMaxBit[ch]));
        paRecoValue[ch][uiYIdxRaster] = (paPixelValue[ch][uiYIdxRaster] * g_invQuantScales[iQPrem[ch]] + iAdd[ch]) >> InvquantiserRightShift[ch];
        paRecoValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[ch][uiYIdxRaster], g_bitDepth[ch?1:0]));
#else
        paPixelValue[ch] [uiYIdxRaster] =  (pOrg[ch][uiYIdxRaster]   *quantiserScale + rightShiftOffset)>>quantiserRightShift;
        paRecoValue[ch][uiYIdxRaster] = (paPixelValue[ch] [uiYIdxRaster]*g_invQuantScales[iQPrem] + iAdd)>>InvquantiserRightShift;
        paRecoValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[ch][uiYIdxRaster], g_bitDepth[ch?1:0]));
#endif
      }
    }

    return false;
  }
  else
  {

    if (bLossless)
    {
      for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
      {
        paPixelValue[ch] [uiYIdxRaster] =  pOrg[ch][uiYIdxRaster]    - uiPredTop[ch];
        paRecoValue[ch][uiYIdxRaster] = pOrg[ch][uiYIdxRaster];
      }
    }
    else
    {

      Int iError, iSign;
      UInt uiLevel, uiQLevel;

      for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
      {
        iError = pOrg[ch][uiYIdxRaster]    - uiPredTop[ch];
        iSign = iError<0? -1: 1;
        uiLevel = iSign*iError;
#if PLT_ESCAPE_ADAPTIVE_BD_FLC
        uiQLevel = Pel(ClipBD<Int>((uiLevel *quantiserScale[ch] + rightShiftOffset[ch]) >> quantiserRightShift[ch], uiMaxBit[ch]));
        paPixelValue[ch][uiYIdxRaster] = iSign * uiQLevel;
        paRecoValue[ch][uiYIdxRaster] =  iSign * ((uiQLevel*g_invQuantScales[iQPrem[ch]] + iAdd[ch])>>InvquantiserRightShift[ch]);
        paRecoValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[ch][uiYIdxRaster] + uiPredTop[ch], g_bitDepth[ch? 1: 0]));
#else
        uiQLevel = (uiLevel   *quantiserScale + rightShiftOffset)>>quantiserRightShift;
        paPixelValue[ch] [uiYIdxRaster] = iSign * uiQLevel;
        paRecoValue[ch][uiYIdxRaster] =  iSign * ((uiQLevel*g_invQuantScales[iQPrem] + iAdd)>>InvquantiserRightShift);
        paRecoValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[ch][uiYIdxRaster] + uiPredTop[ch], g_bitDepth[ch? 1: 0]));
#endif
      }
    }

    return true;
  }
#endif
}

Bool  TComPrediction::CalCopyLuma(TComDataCU* pcCU, Pel* pOrg, Pel *pPalette, Pel* pValue, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
#if PLT_PLT_ERR_LIMIT
  UInt uiErrorThres = bLossless ? 0 : getPLTErrLimit();
#else
  UInt uiErrorThres = bLossless? 0:3; 
#endif
  UInt uiIdx = uiStartPos;
  UInt uiY=0, uiX=0;
  UInt uiBestIdx, uiMinError;
  Bool valid = false;

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx/uiWidth;
    uiX = uiIdx%uiWidth;
    if (uiY == 0)
    {
      return false;
    }

    UInt uiYIdxRaster    = uiY*uiStrideOrg + uiX;
    UInt uiPLTIdx = 0;

    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    {
      UInt uiAbsError = abs (pPalette[uiPLTIdx] - pOrg[uiYIdxRaster]) ;

      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx ++;
    }

    UInt uiStride = uiWidth;


    Bool bMismatch = uiMinError>uiErrorThres;
    if (bMismatch && uiRun == 0)
      return false;

    if (uiRun == 0 && bMismatch)
      return false;

    if (  pValue[uiIdx]==pValue[uiIdx-uiStride] &&pEscapeFlag[uiIdx-uiStride] == 0  && bMismatch == 0)
    {
      uiRun ++;
      valid = true;
    }
    else
    {
      break;
    }

    uiIdx++;
  }
  return valid;
}

 Bool TComPrediction:: CalCopyChroma(TComDataCU* pcCU, Pel* pOrg [2], Pel *pPalette [2],  Pel* pValue, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
 {
  Bool bLossless = pcCU->getCUTransquantBypass(0);
#if PLT_PLT_ERR_LIMIT
  UInt uiErrorThres = bLossless ? 0 : 2 * getPLTErrLimit();
#else
  UInt uiErrorThres = bLossless? 0:3; 
#endif
  UInt uiIdx = uiStartPos;
  UInt uiY=0, uiX=0;
  UInt uiBestIdx, uiMinError;
  Bool valid = false;

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx/uiWidth;
    uiX = uiIdx%uiWidth;
    if (uiY == 0)
    {
      return false;
    }


    UInt uiYIdxRaster    = uiY*uiStrideOrg + uiX;
    UInt uiPLTIdx = 0;

    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    {
      UInt uiAbsError = abs (pPalette[0][uiPLTIdx] - pOrg[0][uiYIdxRaster]) + abs (pPalette[1][uiPLTIdx] - pOrg[1][uiYIdxRaster]) ;
   
      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx ++;
    }

    UInt uiStride = 0;

    uiStride = uiWidth ;
    Bool bMismatch = uiMinError>uiErrorThres;
    if (bMismatch && uiRun == 0)
      return false;

    if (uiRun == 0 && bMismatch)
      return false;

    if (  pValue[uiIdx]==pValue[uiIdx-uiStride] &&pEscapeFlag[uiIdx-uiStride] == 0  && bMismatch == 0)
    {
      uiRun ++;
      valid = true;
    }
    else
    {
      break;
    }

    uiIdx++;
  }
  return valid;

 }

#if PLT_LAST_MERGE || PLT_DERIVATION_MEAN
Bool  TComPrediction::CalCopy(TComDataCU* pcCU, Pel* pOrg [3], Pel *pPalette[3],  Pel* pValue, Pel *pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
	UInt uiIdx = uiStartPos;
	Bool valid = false;

	if(uiIdx < uiWidth)
		return valid;

	while (uiIdx < uiTotal)
	{
		if(pEscapeFlag[uiIdx] || pEscapeFlag[uiIdx-uiWidth])
			break;

		if (  pValue[uiIdx]==pValue[uiIdx-uiWidth])
		{
			uiRun ++;
			valid = true;
		}
		else
		{
			break;
		}

		uiIdx++;
	}
	return valid;
}
#else
Bool  TComPrediction::CalCopy(TComDataCU* pcCU, Pel* pOrg [3], Pel *pPalette[3],  Pel* pValue, Pel *pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);

  UInt uiYCbCrShift = pcCU->getPic()->getComponentScaleY(COMPONENT_Cb);
  UInt uiXCbCrShift = pcCU->getPic()->getComponentScaleX(COMPONENT_Cb);

  UInt uiIdx = uiStartPos;
  UInt uiY=0, uiX=0, uiYCbCr = 0, uiXCbCr = 0;
  UInt uiStrideOrgCbCr = uiStrideOrg >> uiXCbCrShift;
  UInt uiBestIdx, uiMinError;
  ChromaFormat cCF = pcCU->getPic()->getSlice(0)->getSPS()->getChromaFormatIdc();
  Bool testC = true;
  Bool valid = false;
#if PLT_PLT_ERR_LIMIT
  Int iPLTErrLimit = bLossless ? 0 : getPLTErrLimit();
#endif

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    uiX = uiIdx % uiWidth;
    if (uiY == 0)
    {
      return false;
    }

    testC = (((cCF == CHROMA_420) && (uiY % 2 || uiX % 2)) 
          || ((cCF == CHROMA_422) && (uiX % 2))) ? false : true;
    uiYCbCr = uiY >> uiYCbCrShift;
    uiXCbCr = uiX >> uiXCbCrShift;
    UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;
    UInt uiCbCrIdxRaster = uiYCbCr * uiStrideOrgCbCr + uiXCbCr;
    UInt uiPLTIdx = 0;

    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    {
      UInt uiAbsError = abs (pPalette[0][uiPLTIdx] - pOrg[0][uiYIdxRaster]);

      if (testC)
      {
        uiAbsError = uiAbsError+ abs (pPalette[1][uiPLTIdx] - pOrg[1][uiCbCrIdxRaster]) + abs(pPalette[2][uiPLTIdx] - pOrg[2][uiCbCrIdxRaster]);
      }
      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx ++;
    }

    UInt uiStride = 0;

    uiStride = uiWidth ;
    Bool bMismatch; //lossless only
    if (bLossless)
    {
      bMismatch = uiMinError>0;
    }
    else
    {
#if PLT_PLT_ERR_LIMIT
      if (testC)
      {
        bMismatch = uiMinError > 3 * iPLTErrLimit;
      }
      else
      {
        bMismatch = uiMinError > iPLTErrLimit;
      }
#else
      bMismatch = uiMinError>3;
#endif
    }

    if (uiRun == 0 && bMismatch)
    {
      return false;
    }
    if ((pValue[uiIdx] == pValue[uiIdx - uiStride]) && (pEscapeFlag[uiIdx - uiStride] == 0) && (bMismatch == 0))
    {
      uiRun ++;
      valid = true;
    }
    else
    {
      break;
    }

    uiIdx++;
  }
  return valid;
}
#endif
Bool TComPrediction::CalRunLuma (TComDataCU* pcCU, Pel* pOrg, Pel *pPalette, Pel* pValue , Pel * pSPoint, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
#if PLT_PLT_ERR_LIMIT
  UInt uiErrorThres = bLossless ? 0 : getPLTErrLimit();
#else
  UInt uiErrorThres = bLossless? 0: 3;
#endif
  UInt uiIdx = uiStartPos;
  UInt uiY=0, uiX=0;
  UInt uiBestIdx, uiMinError;

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    uiX = uiIdx % uiWidth;
    UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;

    UInt uiPLTIdx = 0;
    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    { 
      UInt uiAbsError = abs (pPalette[uiPLTIdx] - pOrg[uiYIdxRaster]);

      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx++;
    }

    Bool bMismatch = uiMinError > uiErrorThres;
    if (bMismatch && uiIdx == uiStartPos)
    {
      return false;
    }

    if ( (uiIdx && (pSPoint[uiIdx-1]==0 && pEscapeFlag[uiIdx-1]==0)&&pValue[uiIdx]==pValue[uiIdx-1]) && bMismatch == 0)
    {
      uiRun ++;
    }
    else if (uiIdx > uiStartPos)
    {
      break;
    }
    uiIdx++;
  }
  return true;
}


Bool TComPrediction::CalRunChroma(TComDataCU* pcCU, Pel* pOrg [2], Pel *pPalette [2], Pel* pValue , Pel * pSPoint, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
#if PLT_PLT_ERR_LIMIT
  UInt uiErrorThres = bLossless ? 0 : 2 * getPLTErrLimit();
#else
  UInt uiErrorThres = bLossless? 0: 3;
#endif
  UInt uiIdx = uiStartPos;
  UInt uiY = 0, uiX = 0;
  UInt uiBestIdx, uiMinError;

  while (uiIdx < uiTotal)
  {

    uiY = uiIdx / uiWidth;
    uiX = uiIdx % uiWidth;

    UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;
    UInt uiPLTIdx = 0;
    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    { 

      UInt uiAbsError = abs(pPalette[0][uiPLTIdx] - pOrg[0][uiYIdxRaster]) + abs(pPalette[1][uiPLTIdx] - pOrg[1][uiYIdxRaster]);

      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx ++;
    }

    Bool bMismatch = uiMinError > uiErrorThres;
    if (bMismatch && uiIdx == uiStartPos)
    {
      return false;
    }

    if (uiIdx && (pSPoint[uiIdx - 1] == 0) && (pEscapeFlag[uiIdx - 1] == 0) 
              && (pValue[uiIdx] == pValue[uiIdx - 1]) && (bMismatch == 0))
    {
      uiRun ++;
    }
    else if (uiIdx > uiStartPos)
    {
      break;
    }
    uiIdx++;
  }
  return true;

}
#if PLT_LAST_MERGE || PLT_DERIVATION_MEAN
Bool TComPrediction::CalRun(TComDataCU* pcCU, Pel* pOrg [3], Pel *pPalette [3] , Pel* pValue , Pel * pSPoint, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
	UInt uiIdx = uiStartPos;

	while (uiIdx < uiTotal)
	{
		if(pEscapeFlag[uiIdx])
			break;

		if ( uiIdx && pSPoint[uiIdx-1]==0 && pEscapeFlag[uiIdx-1]==0 && pValue[uiIdx]==pValue[uiIdx-1])
		{
			uiRun ++;
		}
		else if (uiIdx> uiStartPos)
		{
			break;
		}
		uiIdx++;
	}
	return (!(pEscapeFlag[uiIdx] && uiIdx==uiStartPos));
}
#else
Bool TComPrediction::CalRun(TComDataCU* pcCU, Pel* pOrg [3], Pel *pPalette [3] , Pel* pValue , Pel * pSPoint, Pel * pEscapeFlag, Pel* pRun, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  UInt uiYCbCrShift = pcCU->getPic()->getComponentScaleY(COMPONENT_Cb);
  UInt uiXCbCrShift = pcCU->getPic()->getComponentScaleX(COMPONENT_Cb);

  UInt uiIdx = uiStartPos;
  UInt uiY=0, uiX=0, uiYCbCr = 0, uiXCbCr = 0;
  UInt uiStrideOrgCbCr = uiStrideOrg >> uiXCbCrShift;
  UInt uiBestIdx, uiMinError;
  ChromaFormat cCF = pcCU->getPic()->getSlice(0)->getSPS()->getChromaFormatIdc();
  Bool testC = true;
#if PLT_PLT_ERR_LIMIT
  Int iPLTErrLimit = bLossless? 0: getPLTErrLimit();
#endif

  while (uiIdx < uiTotal)
  {
    uiY = uiIdx / uiWidth;
    uiX = uiIdx % uiWidth;

    uiYCbCr = uiY >> uiYCbCrShift;
    uiXCbCr = uiX >> uiXCbCrShift;
    UInt uiYIdxRaster = uiY * uiStrideOrg + uiX;
    UInt uiCbCrIdxRaster = uiYCbCr * uiStrideOrgCbCr + uiXCbCr;
    UInt uiPLTIdx = 0;
    testC = (((cCF == CHROMA_420) && (uiY % 2 || uiX % 2)) || 
             ((cCF == CHROMA_422) && (uiX % 2))) ? false : true;
    uiMinError = MAX_UINT;
    while (uiPLTIdx < uiPLTSize)
    {
      UInt uiAbsError = abs (pPalette[0][uiPLTIdx] - pOrg[0][uiYIdxRaster]);
      if (testC)
      {
        uiAbsError = uiAbsError+ abs (pPalette[1][uiPLTIdx] - pOrg[1][uiCbCrIdxRaster]) + abs(pPalette[2][uiPLTIdx] - pOrg[2][uiCbCrIdxRaster]);
      }
      if (uiAbsError < uiMinError)
      {
        uiBestIdx = uiPLTIdx;
        uiMinError = uiAbsError;
        pValue[uiIdx] = uiBestIdx;
        if (uiMinError == 0)
        {
          break;
        }
      }
      uiPLTIdx ++;
    }

    Bool bMismatch;
    if (bLossless)
    {
      bMismatch = uiMinError>0;
    }
    else
    {
#if PLT_PLT_ERR_LIMIT
      if (testC)
      {
        bMismatch = uiMinError> 3 * iPLTErrLimit;
      }
      else
      {
        bMismatch = uiMinError> iPLTErrLimit;
      }
#else
      bMismatch = uiMinError > 3;
#endif
    }
    if (bMismatch && uiIdx == uiStartPos)
    {
      return false;
    }

    if ( (uiIdx && (pSPoint[uiIdx-1]==0 && pEscapeFlag[uiIdx-1]==0)&&pValue[uiIdx]==pValue[uiIdx-1]) && bMismatch == 0)
    {
      uiRun ++;
    }
    else if (uiIdx> uiStartPos)
    {
      break;
    }
    uiIdx++;
  }
  return true;
}
#endif
#endif
//! \}
